// Copyright (c) 2002 Tom Barry.  All rights reserved.
//      trbarry@trbarry.com
// Requires Avisynth source code to compile for Avisynth
// Avisynth Copyright 2000 Ben Rudiak-Gould.
//      http://www.math.berkeley.edu/~benrg/avisynth.html
/////////////////////////////////////////////////////////////////////////////
//
//	This file is subject to the terms of the GNU General Public License as
//	published by the Free Software Foundation.  A copy of this license is
//	included with this software distribution in the file COPYING.  If you
//	do not have a copy, you may obtain a copy by writing to the Free
//	Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
//	This software is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details
//  
//  Also, this program is "Philanthropy-Ware".  That is, if you like it and 
//  feel the need to reward or inspire the author then please feel free (but
//  not obligated) to consider joining or donating to the Electronic Frontier
//  Foundation. This will help keep cyber space free of barbed wire and bullsh*t.  
//
/////////////////////////////////////////////////////////////////////////////
// Change Log
//
// Date          Version      Developer             Changes
//
// 30 Jan 2003   0.0.1.3      Tom Barry		        PlugInit2, VS6 support 

// 2? Nov 2002   0.0.1.2      SansGrip		        YUY2 Support 

// 19 Nov 2002   0.0.1.0      Tom Barry		        Create DctFilter 
//
/////////////////////////////////////////////////////////////////////////////
// CVS Log
//
//
/////////////////////////////////////////////////////////////////////////////

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <crtdbg.h>

#ifdef AVISYNTH_20
#include "avisynth20.h"
#else
#include "avisynth.h"
#endif

#include <malloc.h>
#include <string.h>
#include "DctFilter.h"


const __int64 ZEROS = 0x0000000000000000;

class DctFilter : public GenericVideoFilter {
    int  Max_Frames;

    PVideoFrame src;
    PVideoFrame dst;

    int src_pitch;
    int dst_pitch;
    int rowsize;
    int height;
    int FldHeight;
	bool SSE2enabled;
	bool SSEMMXenabled;
	bool _3DNOWenabled;
	short * pWorkArea, *pWorkArea_u, *pWorkArea_v;
	fdctFuncPtr fdct;
	idctFuncPtr idct;

	double fac[8];
	short factors[8][8];

#ifndef AVISYNTH_20
	int __stdcall Fieldcopy(void *dest, const void *src, size_t count, 
				int rows, int dst_pitch, int src_pitch);
	BOOL GoFilterIt(int src_pit, int dst_pit,
			int rowsize, const BYTE* srcp, BYTE* dstp, int FldHeight);
	PVideoFrame GetFrame_YV12(int inFrame, IScriptEnvironment* env);
	PVideoFrame GetFrame_YUY2(int inFrame, IScriptEnvironment* env);
#endif

public:
    DctFilter(PClip _child, 
		double _fac0, double _fac1, double _fac2, double _fac3, 
		double _fac4, double _fac5, double _fac6, double _fac7,
		IScriptEnvironment* env) : GenericVideoFilter(_child)
    {
		_asm emms;

        Max_Frames = vi.num_frames;
		SSEMMXenabled = (env->GetCPUFlags() & CPUF_INTEGER_SSE) !=0;
		SSE2enabled = (env->GetCPUFlags() & CPUF_SSE2) !=0;

		if ((vi.width % 16) || (vi.height % 16))
		{
			env->ThrowError("Sorry, DctFiler test version needs size multiples of 16x16");
		}

#ifdef AVISYNTH_20
		if(!vi.IsYUY2())
			env->ThrowError("DctFilter: Supports YUY2 color format only");
#else
	    if (!vi.IsYV12() && !vi.IsYUY2())
			env->ThrowError("DctFilter: Supports YV12 and YUY2 color formats only");
#endif
	    
		if (SSE2enabled)
		{
//>>>			fdct = fdct_sse2;		//		fdct(&data[i*64]);
//			idct = idct_sse2;
			fdct = fdct_mmx;		
			idct = idct_xmm;
		}
	    else if (SSEMMXenabled)
		{
			fdct = fdct_mmx;		
			idct = idct_xmm;
		}
		else
		{
			fdct = fdct_mmx;		
			idct = idct_mmx;
		}

		pWorkArea = (short * const) _aligned_malloc(8*8*2, 128);
		if(vi.IsYUY2())
		{
			pWorkArea_u = (short * const) _aligned_malloc(8*8*2, 128);
			pWorkArea_v = (short * const) _aligned_malloc(8*8*2, 128);
		}

		fac[0] = _fac0;
		fac[1] = _fac1;
		fac[2] = _fac2;
		fac[3] = _fac3;
		fac[4] = _fac4;
		fac[5] = _fac5;
		fac[6] = _fac6;
		fac[7] = _fac7;

		for (int i=0; i <=7; i++)
		{
			for (int j=0; j <=7; j++)
			{
				factors[i][j] = (short) (fac[i] * fac[j] * 8);
			}
		}
    }

    PVideoFrame __stdcall GetFrame(int inFrame, IScriptEnvironment* env);
};

#ifndef AVISYNTH_20

PVideoFrame __stdcall DctFilter::GetFrame(int inFrame, IScriptEnvironment* env)
{
	if(vi.IsYUY2())
		return GetFrame_YUY2(inFrame, env);
	else
		return GetFrame_YV12(inFrame, env);

	// Unreachable

	return 0;
}

PVideoFrame DctFilter::GetFrame_YV12(int inFrame, IScriptEnvironment* env) 
{

const BYTE* srcp;
unsigned char* dstp;
	

    src = child->GetFrame(inFrame, env);
	srcp = src->GetReadPtr();

	if (inFrame < 0 || inFrame >= Max_Frames)
	{
		return src;						// don't filter first and last frames
	}

    dst = env->NewVideoFrame(vi);

	srcp = src->GetReadPtr(PLANAR_Y);
	dstp = dst->GetWritePtr(PLANAR_Y);
	src_pitch = src->GetPitch(PLANAR_Y);
	dst_pitch = dst->GetPitch(PLANAR_Y);
	rowsize = dst->GetRowSize(PLANAR_Y);   // Could also be PLANAR_Y_ALIGNED which would return a mod16 row_size   
	height = dst->GetHeight(PLANAR_Y);

	GoFilterIt(src_pitch, dst_pitch,
		rowsize, srcp, dstp, height);					// go H expand lines into even output lines

		// Next, the U plane
	srcp = src->GetReadPtr(PLANAR_U);
	dstp = dst->GetWritePtr(PLANAR_U);
	src_pitch = src->GetPitch(PLANAR_U);
	dst_pitch = dst->GetPitch(PLANAR_U);
	rowsize = dst->GetRowSize(PLANAR_U);  // Could also be PLANAR_U_ALIGNED which would return a mod8 row_size   
	height = dst->GetHeight(PLANAR_U);

	GoFilterIt(src_pitch, dst_pitch,
		rowsize, srcp, dstp, height);					// go H expand lines into even output lines

	// And the V plane, same sizes as U, different locations
	srcp = src->GetReadPtr(PLANAR_V);   
	dstp = dst->GetWritePtr(PLANAR_V);   

	GoFilterIt(src_pitch, dst_pitch,
		rowsize, srcp, dstp, height);					// go H expand lines into even output lines

	__asm emms;

    return dst;
}

#define IS_SSE
#define SSE_TYPE SSE

#pragma warning(disable: 4799) // "no emms"

BOOL	DctFilter::GoFilterIt(int src_pit, int dst_pit,
			int rowsize, const BYTE* srcp, BYTE* dstp, 	int FldHeight)					// go H expand lines into even output lines
{
const BYTE* pSrc = srcp;
const BYTE* pSrcW;
BYTE* pDest = dstp;
BYTE* pDestW;

int y;
int ct = (rowsize/8);
int ctW;
short* pFactors = &factors[0][0];
short* pWorkAreaW = pWorkArea;			// so inline asm can access

#include "LoopTopYV12.inc"
		
		fdct(pWorkAreaW);					// go do forward DCT	
		// lets adjust some of the DCT components
		DO_ADJUST(pWorkAreaW);
		idct(pWorkAreaW);					// go do IDCT and reverse it

#include "LoopBottomYV12.inc"

	return 0;
}


#pragma warning(default: 4799)

PVideoFrame DctFilter::GetFrame_YUY2(int inFrame, IScriptEnvironment* env)
#else // AVISYNTH_20
PVideoFrame DctFilter::GetFrame(int inFrame, IScriptEnvironment* env)
#endif
{
	short* pWorkAreaW = pWorkArea;			// so inline asm can access
	short* pWorkArea_uW = pWorkArea_u;			// so inline asm can access
	short* pWorkArea_vW = pWorkArea_v;			// so inline asm can access
	fdctFuncPtr fdctW = fdct;					// "
	idctFuncPtr idctW = idct;

	PVideoFrame f = child->GetFrame(inFrame, env);
	const BYTE* srcp = f->GetReadPtr();

	const int row_size = f->GetRowSize(), src_pitch = f->GetPitch(),
		src_pitch_x8 = src_pitch << 3;
	int ycnt = f->GetHeight() >> 3;

	f = env->NewVideoFrame(vi);
	BYTE* dstp = f->GetWritePtr();

	const int dst_pitch = f->GetPitch(), dst_pitch_x8 = dst_pitch << 3;
	const __int64 y_mask = 0x00ff00ff00ff00ff,
		u_mask = 0x0000ff000000ff00,
		v_mask = 0xff000000ff000000;

	short* pFactors = &factors[0][0];
	
#include "LoopTopYUY2.inc"

		fdct(pWorkAreaW);
		DO_ADJUST(pWorkAreaW);
		idct(pWorkAreaW);					// go do IDCT and reverse it

		fdct(pWorkArea_uW);
		DO_ADJUST(pWorkArea_uW);
		idct(pWorkArea_uW);					// go do IDCT and reverse it
		
		fdct(pWorkArea_vW);
		DO_ADJUST(pWorkArea_vW);
		idct(pWorkArea_vW);					// go do IDCT and reverse it

#include "LoopBottomYUY2.inc" 				// Pels back to frame
		
		return f;
}

//--------------------- DctFilterD ------------------------------------------------------------

class DctFilterD : public GenericVideoFilter {
    int  Max_Frames;

    PVideoFrame src;
    PVideoFrame dst;

    int src_pitch;
    int dst_pitch;
    int rowsize;
    int height;
    int FldHeight;
	bool SSE2enabled;
	bool SSEMMXenabled;
	bool _3DNOWenabled;
	short * pWorkArea, *pWorkArea_u, *pWorkArea_v;
	fdctFuncPtr fdct;
	idctFuncPtr idct;

	short factors[8][8];

#ifndef AVISYNTH_20
	int __stdcall Fieldcopy(void *dest, const void *src, size_t count, 
				int rows, int dst_pitch, int src_pitch);
	BOOL GoFilterIt(int src_pit, int dst_pit,
			int rowsize, const BYTE* srcp, BYTE* dstp, int FldHeight);
	PVideoFrame GetFrame_YV12(int inFrame, IScriptEnvironment* env);
	PVideoFrame GetFrame_YUY2(int inFrame, IScriptEnvironment* env);
#endif

public:
    DctFilterD(PClip _child, 
		int DiagCt, IScriptEnvironment* env) : GenericVideoFilter(_child)
    {
		_asm emms;

        Max_Frames = vi.num_frames;
		SSEMMXenabled = (env->GetCPUFlags() & CPUF_INTEGER_SSE) !=0;
		SSE2enabled = (env->GetCPUFlags() & CPUF_SSE2) !=0;

		if ((vi.width % 16) || (vi.height % 16))
		{
			env->ThrowError("Sorry, DctFilter test version needs size multiples of 16x16");
		}
		if (DiagCt > 14 || DiagCt < 1)
		{
			env->ThrowError("DctFilter Diag Count must be 1-14");
		}

#ifdef AVISYNTH_20
		if(!vi.IsYUY2())
			env->ThrowError("DctFilterD: Supports YUY2 color format only");
#else
	    if (!vi.IsYV12() && !vi.IsYUY2())
			env->ThrowError("DctFilterD: Supports YV12 and YUY2 color formats only");
#endif
	    
		if (SSE2enabled)
		{
//>>>			fdct = fdct_sse2;		//		fdct(&data[i*64]);
//			idct = idct_sse2;
			fdct = fdct_mmx;		
			idct = idct_xmm;
		}
	    else if (SSEMMXenabled)
		{
			fdct = fdct_mmx;		
			idct = idct_xmm;
		}
		else
		{
			fdct = fdct_mmx;		
			idct = idct_mmx;
		}

		pWorkArea = (short * const) _aligned_malloc(8*8*2, 128);
		if(vi.IsYUY2())
		{
			pWorkArea_u = (short * const) _aligned_malloc(8*8*2, 128);
			pWorkArea_v = (short * const) _aligned_malloc(8*8*2, 128);
		}


		for (int i=0; i <=7; i++)
		{
			for (int j=0; j <=7; j++)
			{
				factors[i][j] = (i+j > 14 - DiagCt)
					? 0
					: 0xffff;
			}
		}
    }

    PVideoFrame __stdcall GetFrame(int inFrame, IScriptEnvironment* env);
};

#ifndef AVISYNTH_20

PVideoFrame __stdcall DctFilterD::GetFrame(int inFrame, IScriptEnvironment* env)
{
	if(vi.IsYUY2())
		return GetFrame_YUY2(inFrame, env);
	else
		return GetFrame_YV12(inFrame, env);

	// Unreachable

	return 0;
}

PVideoFrame DctFilterD::GetFrame_YV12(int inFrame, IScriptEnvironment* env) 
{

const BYTE* srcp;
unsigned char* dstp;
	

    src = child->GetFrame(inFrame, env);
	srcp = src->GetReadPtr();

	if (inFrame < 0 || inFrame >= Max_Frames)
	{
		return src;						// don't filter first and last frames
	}

    dst = env->NewVideoFrame(vi);

	srcp = src->GetReadPtr(PLANAR_Y);
	dstp = dst->GetWritePtr(PLANAR_Y);
	src_pitch = src->GetPitch(PLANAR_Y);
	dst_pitch = dst->GetPitch(PLANAR_Y);
	rowsize = dst->GetRowSize(PLANAR_Y);   // Could also be PLANAR_Y_ALIGNED which would return a mod16 row_size   
	height = dst->GetHeight(PLANAR_Y);

	GoFilterIt(src_pitch, dst_pitch,
		rowsize, srcp, dstp, height);					// go H expand lines into even output lines

		// Next, the U plane
	srcp = src->GetReadPtr(PLANAR_U);
	dstp = dst->GetWritePtr(PLANAR_U);
	src_pitch = src->GetPitch(PLANAR_U);
	dst_pitch = dst->GetPitch(PLANAR_U);
	rowsize = dst->GetRowSize(PLANAR_U);  // Could also be PLANAR_U_ALIGNED which would return a mod8 row_size   
	height = dst->GetHeight(PLANAR_U);

	GoFilterIt(src_pitch, dst_pitch,
		rowsize, srcp, dstp, height);					// go H expand lines into even output lines

	// And the V plane, same sizes as U, different locations
	srcp = src->GetReadPtr(PLANAR_V);   
	dstp = dst->GetWritePtr(PLANAR_V);   

	GoFilterIt(src_pitch, dst_pitch,
		rowsize, srcp, dstp, height);					// go H expand lines into even output lines

	__asm emms;

    return dst;
}

#define IS_SSE
#define SSE_TYPE SSE

#pragma warning(disable: 4799) // "no emms"

BOOL	DctFilterD::GoFilterIt(int src_pit, int dst_pit,
			int rowsize, const BYTE* srcp, BYTE* dstp, 	int FldHeight)					// go H expand lines into even output lines
{
const BYTE* pSrc = srcp;
const BYTE* pSrcW;
BYTE* pDest = dstp;
BYTE* pDestW;

int y;
int ct = (rowsize/8);
int ctW;
short* pFactors = &factors[0][0];
short* pWorkAreaW = pWorkArea;			// so inline asm can access

#include "LoopTopYV12.inc"
		
		fdct(pWorkAreaW);					// go do forward DCT	
		// lets adjust some of the DCT components
		DO_ADJUSTD(pWorkAreaW);
		idct(pWorkAreaW);					// go do IDCT and reverse it

#include "LoopBottomYV12.inc"

	return 0;
}


#pragma warning(default: 4799)

PVideoFrame DctFilterD::GetFrame_YUY2(int inFrame, IScriptEnvironment* env)
#else // AVISYNTH_20
PVideoFrame DctFilterD::GetFrame(int inFrame, IScriptEnvironment* env)
#endif
{
	short* pWorkAreaW = pWorkArea;			// so inline asm can access
	short* pWorkArea_uW = pWorkArea_u;			// so inline asm can access
	short* pWorkArea_vW = pWorkArea_v;			// so inline asm can access
	fdctFuncPtr fdctW = fdct;					// "
	idctFuncPtr idctW = idct;

	PVideoFrame f = child->GetFrame(inFrame, env);
	const BYTE* srcp = f->GetReadPtr();

	const int row_size = f->GetRowSize(), src_pitch = f->GetPitch(),
		src_pitch_x8 = src_pitch << 3;
	int ycnt = f->GetHeight() >> 3;

	f = env->NewVideoFrame(vi);
	BYTE* dstp = f->GetWritePtr();

	const int dst_pitch = f->GetPitch(), dst_pitch_x8 = dst_pitch << 3;
	const __int64 y_mask = 0x00ff00ff00ff00ff,
		u_mask = 0x0000ff000000ff00,
		v_mask = 0xff000000ff000000;

	short* pFactors = &factors[0][0];
	
#include "LoopTopYUY2.inc"

		fdct(pWorkAreaW);
		DO_ADJUSTD(pWorkAreaW);
		idct(pWorkAreaW);					// go do IDCT and reverse it

		fdct(pWorkArea_uW);
		DO_ADJUSTD(pWorkArea_uW);
		idct(pWorkArea_uW);					// go do IDCT and reverse it
		
		fdct(pWorkArea_vW);
		DO_ADJUSTD(pWorkArea_vW);
		idct(pWorkArea_vW);					// go do IDCT and reverse it

#include "LoopBottomYUY2.inc" 				// Pels back to frame
		
		return f;
}




//--------------------- DctAddConstant ---------------------------------------------------------

class DctAddConstant : public GenericVideoFilter {
    int  Max_Frames;

    PVideoFrame src;
    PVideoFrame dst;

    int src_pitch;
    int dst_pitch;
    int rowsize;
    int height;
    int FldHeight;
	bool SSE2enabled;
	bool SSEMMXenabled;
	bool _3DNOWenabled;
	short * pWorkArea, *pWorkArea_u, *pWorkArea_v;
	fdctFuncPtr fdct;
	idctFuncPtr idct;

	double fac[8];
	short factors[8][4];

#ifndef AVISYNTH_20
	int __stdcall Fieldcopy(void *dest, const void *src, size_t count, 
				int rows, int dst_pitch, int src_pitch);
	BOOL GoFilterIt(int src_pit, int dst_pit,
			int rowsize, const BYTE* srcp, BYTE* dstp, int FldHeight);
	PVideoFrame GetFrame_YV12(int inFrame, IScriptEnvironment* env);
	PVideoFrame GetFrame_YUY2(int inFrame, IScriptEnvironment* env);
#endif

public:
    DctAddConstant(PClip _child, 
		double _fac4, double _fac5, double _fac6, double _fac7,
		IScriptEnvironment* env) : GenericVideoFilter(_child)
    {
		_asm emms;

        Max_Frames = vi.num_frames;
		SSEMMXenabled = (env->GetCPUFlags() & CPUF_INTEGER_SSE) !=0;
		SSE2enabled = (env->GetCPUFlags() & CPUF_SSE2) !=0;

		if ((vi.width % 16) || (vi.height % 16))
		{
			env->ThrowError("Sorry, DctFiler test version needs size multiples of 16x16");
		}

#ifdef AVISYNTH_20
		if(!vi.IsYUY2())
			env->ThrowError("DctAddConstant: Supports YUY2 color format only");
#else
	    if (!vi.IsYV12() && !vi.IsYUY2())
			env->ThrowError("DctAddConstant: Supports YV12 and YUY2 color formats only");
#endif
	    
		if (SSE2enabled)
		{
			fdct = fdct_mmx;		
			idct = idct_xmm;
		}
	    else if (SSEMMXenabled)
		{
			fdct = fdct_mmx;		
			idct = idct_xmm;
		}
		else
		{
			fdct = fdct_mmx;		
			idct = idct_mmx;
		}

		pWorkArea = (short * const) _aligned_malloc(8*8*2, 128);
		if(vi.IsYUY2())
		{
			pWorkArea_u = (short * const) _aligned_malloc(8*8*2, 128);
			pWorkArea_v = (short * const) _aligned_malloc(8*8*2, 128);
		}

		fac[0] = 0;
		fac[1] = 0;
		fac[2] = 0;
		fac[3] = 0;
		fac[4] = _fac4;
		fac[5] = _fac5;
		fac[6] = _fac6;
		fac[7] = _fac7;

		for (int j=0; j <=3; j++)
		{
			factors[0][j] = (short) (fac[j+4] * 8);
		}

		for (int i=4; i <=7; i++)
		{
			for (int j=0; j <=3; j++)
			{
				factors[i][j] = (short) (fac[i] * 8);
			}
		}
    }

    PVideoFrame __stdcall GetFrame(int inFrame, IScriptEnvironment* env);
};

#ifndef AVISYNTH_20

PVideoFrame __stdcall DctAddConstant::GetFrame(int inFrame, IScriptEnvironment* env)
{
	if(vi.IsYUY2())
		return GetFrame_YUY2(inFrame, env);
	else
		return GetFrame_YV12(inFrame, env);

	// Unreachable

	return 0;
}

#define ADD_TEXTURE2(num) \
	__asm { \
	__asm pshufw mm0, qword ptr[eax+num*16], 3<<6 | 3<<4 | 3<<2 | 3	/* all high word */ \
	__asm movq mm1, qword ptr[eax+num*16+8]		/* high word */ \
	__asm pmullw mm0, qword ptr[esi+0*8]	    /* mult by col parm factor */ \
	__asm psraw mm0, 3 \
	__asm paddsw mm1, mm0 						/* add texture amount */ \
	__asm movq qword ptr[eax+0*16+8], mm1 		/* and put high qword back into work */ \
	}


#define ADD_TEXTURE3(num) \
	__asm { \
	__asm movq	mm0, qword ptr[eax+num*16]			/* low qword */ \
	__asm movq	mm1, qword ptr[eax+num*16+8]		/* high word */ \
\
	/* 1st get texture for high word, also needs col factor */ \
	__asm pshufw mm2, mm0, 3<<6 | 3<<4 | 3<<2 | 3	/* all high word of low qword */ \
	__asm pmullw mm2, qword ptr[esi+0*8]		/* mult by col parm factor */ \
	__asm movq	mm4, mm7						/* saved copy of hi qword of row 4 */ \
	__asm pmullw mm4, qword ptr[esi+num*8]	    /* mult by row parm factor */ \
	__asm paddsw mm2, mm4 						/* add texture amounts */ \
	__asm psraw mm2, 3 \
	__asm paddsw mm1, mm2 						/* add texture amount */ \
\
	/* then texture for low qword, needs only row factor (vert texture) */\
	__asm movq	mm4, mm6						/* saved copy of row 4 low qword */ \
	__asm pmullw mm4, qword ptr[esi+num*8]	    /* mult by row parm factor */ \
	__asm psraw mm4, 3 \
	__asm paddsw mm0, mm4 						/* add texture amounts */ \
	__asm movq qword ptr[eax+num*16], mm0 		/* put low qword back into work */ \
	__asm movq qword ptr[eax+num*16+8], mm1 	/* put high qword back into work */ \
	}


#define ADD_TEXTURE(pwa) \
	__asm { \
	__asm mov eax, pwa \
	__asm mov esi, pFactors						/* our parms, mult factors */ \
\
	__asm movq mm6, qword ptr[eax+3*16]			/* low qword of row 3  */ \
	__asm movq mm7, qword ptr[eax+3*16+8]		/* high qword of row 3*/ \
\
	ADD_TEXTURE2(0) \
	ADD_TEXTURE2(1) \
	ADD_TEXTURE2(2) \
	ADD_TEXTURE2(3) \
	ADD_TEXTURE3(4) \
	ADD_TEXTURE3(5) \
	ADD_TEXTURE3(6) \
	ADD_TEXTURE3(7) \
}


PVideoFrame DctAddConstant::GetFrame_YV12(int inFrame, IScriptEnvironment* env) 
{

const BYTE* srcp;
unsigned char* dstp;
	

    src = child->GetFrame(inFrame, env);
	srcp = src->GetReadPtr();

	if (inFrame < 0 || inFrame >= Max_Frames)
	{
		return src;						// don't filter first and last frames
	}

    dst = env->NewVideoFrame(vi);

	srcp = src->GetReadPtr(PLANAR_Y);
	dstp = dst->GetWritePtr(PLANAR_Y);
	src_pitch = src->GetPitch(PLANAR_Y);
	dst_pitch = dst->GetPitch(PLANAR_Y);
	rowsize = dst->GetRowSize(PLANAR_Y);   // Could also be PLANAR_Y_ALIGNED which would return a mod16 row_size   
	height = dst->GetHeight(PLANAR_Y);

	GoFilterIt(src_pitch, dst_pitch,
		rowsize, srcp, dstp, height);					// go H expand lines into even output lines

		// Next, the U plane
	srcp = src->GetReadPtr(PLANAR_U);
	dstp = dst->GetWritePtr(PLANAR_U);
	src_pitch = src->GetPitch(PLANAR_U);
	dst_pitch = dst->GetPitch(PLANAR_U);
	rowsize = dst->GetRowSize(PLANAR_U);  // Could also be PLANAR_U_ALIGNED which would return a mod8 row_size   
	height = dst->GetHeight(PLANAR_U);

	GoFilterIt(src_pitch, dst_pitch,
		rowsize, srcp, dstp, height);					// go H expand lines into even output lines

	// And the V plane, same sizes as U, different locations
	srcp = src->GetReadPtr(PLANAR_V);   
	dstp = dst->GetWritePtr(PLANAR_V);   

	GoFilterIt(src_pitch, dst_pitch,
		rowsize, srcp, dstp, height);					// go H expand lines into even output lines

	__asm emms;

    return dst;
}


#pragma warning(disable: 4799) // "no emms"

BOOL	DctAddConstant::GoFilterIt(int src_pit, int dst_pit,
			int rowsize, const BYTE* srcp, BYTE* dstp, 	int FldHeight)					// go H expand lines into even output lines
{
const BYTE* pSrc = srcp;
const BYTE* pSrcW;
BYTE* pDest = dstp;
BYTE* pDestW;

int y;
int ct = (rowsize/8);
int ctW;
short* pFactors = &factors[0][0];
short* pWorkAreaW = pWorkArea;			// so inline asm can access

#include "LoopTopYV12.inc"
		
		fdct(pWorkAreaW);					// go do forward DCT	
		// lets adjust some of the DCT components
		ADD_TEXTURE(pWorkAreaW);
		idct(pWorkAreaW);					// go do IDCT and reverse it

#include "LoopBottomYV12.inc"

	return 0;
}


#pragma warning(default: 4799)

PVideoFrame DctAddConstant::GetFrame_YUY2(int inFrame, IScriptEnvironment* env)
#else // AVISYNTH_20
PVideoFrame DctAddConstant::GetFrame(int inFrame, IScriptEnvironment* env)
#endif
{
	short* pWorkAreaW = pWorkArea;			// so inline asm can access
	short* pWorkArea_uW = pWorkArea_u;			// so inline asm can access
	short* pWorkArea_vW = pWorkArea_v;			// so inline asm can access
	fdctFuncPtr fdctW = fdct;					// "
	idctFuncPtr idctW = idct;

	PVideoFrame f = child->GetFrame(inFrame, env);
	const BYTE* srcp = f->GetReadPtr();

	const int row_size = f->GetRowSize(), src_pitch = f->GetPitch(),
		src_pitch_x8 = src_pitch << 3;
	int ycnt = f->GetHeight() >> 3;

	f = env->NewVideoFrame(vi);
	BYTE* dstp = f->GetWritePtr();

	const int dst_pitch = f->GetPitch(), dst_pitch_x8 = dst_pitch << 3;
	const __int64 y_mask = 0x00ff00ff00ff00ff,
		u_mask = 0x0000ff000000ff00,
		v_mask = 0xff000000ff000000;

	short* pFactors = &factors[0][0];
	
#include "LoopTopYUY2.inc"

		fdct(pWorkAreaW);
		DO_ADJUST(pWorkAreaW);
		idct(pWorkAreaW);					// go do IDCT and reverse it

		fdct(pWorkArea_uW);
		DO_ADJUST(pWorkArea_uW);
		idct(pWorkArea_uW);					// go do IDCT and reverse it
		
		fdct(pWorkArea_vW);
		DO_ADJUST(pWorkArea_vW);
		idct(pWorkArea_vW);					// go do IDCT and reverse it

#include "LoopBottomYUY2.inc" 				// Pels back to frame
		
		return f;
}


AVSValue __cdecl Create_DctFilter(AVSValue args, void* user_data,
IScriptEnvironment* env) 
{
	_asm emms;
    return new DctFilter(args[0].AsClip(),
		args[1].AsFloat(),
		args[2].AsFloat(),
		args[3].AsFloat(),
		args[4].AsFloat(),
		args[5].AsFloat(),
		args[6].AsFloat(),
		args[7].AsFloat(),
		args[8].AsFloat(),
		env);
}

AVSValue __cdecl Create_DctFilterD(AVSValue args, void* user_data,
IScriptEnvironment* env) 
{
	_asm emms;
    return new DctFilterD(args[0].AsClip(),
		args[1].AsInt(),
		env);
}

AVSValue __cdecl Create_DctAddConstant(AVSValue args, void* user_data,
IScriptEnvironment* env) 
{
	_asm emms;
    return new DctAddConstant(args[0].AsClip(),
		args[1].AsFloat(),
		args[2].AsFloat(),
		args[3].AsFloat(),
		args[4].AsFloat(),
		env);
}

extern "C" __declspec(dllexport) const char* __stdcall
AvisynthPluginInit2(IScriptEnvironment* env) {
	_asm emms;
    env->AddFunction("DctFilter", "cffffffff", Create_DctFilter, 0);
    env->AddFunction("DctFilterD", "ci", Create_DctFilterD, 0);
    env->AddFunction("DctAddConstant", "cffff", Create_DctAddConstant, 0);
    return "`DctFilter' Clip in Space and Time";
}

